# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/91_interface.ipynb.

# %% auto 0
__all__ = ['src_finder', 'defaults', 'proc', 'Summarize', 'SourceAnalyzer', 'SummaryInfo', 'phase_plots', 'harmonic_plots',
           'get_SA', 'examine_source', 'PhaseInfo', 'process_df', 'WTSkyCoord', 'load_source_spreadsheet',
           'get_fermi_info', 'setup_excel', 'process_excel']

# %% ../nbs/91_interface.ipynb 2
import os, sys, warnings
import datetime

from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from utilities.ipynb_docgen import ( show, DataFrameWrapper, ipynb_doc, capture_hide, capture, monospace)
from utilities.catalogs import (UWcat, Fermi4FGL)
from . import (WtLike, PointSource, SourceFinder, Config)
from astropy.coordinates import SkyCoord

# %% ../nbs/91_interface.ipynb 3
"""
Analysis of the time behaviour of a source

To run an individual source: examine_source(name)
where name can be:
* any source identifier recognized by SIMBAD
* the name of a uw source in uw1410, specifically in the wtlike table of weights
* a "j-name": format J1234+5678 or J1234.5-6789
"""
_uwname='uw1410'
plt.rc('font', size=12)

def _float_format(f):
    if f==0: return '0'
    return f'{f:.3f}' if (abs(f)>1e-3 and abs(f)<1e5 ) else f'{f:.2e}'

pd.set_option('display.float_format', _float_format)

with capture_hide(f'Catalog setup: {_uwname} and 4FGL-DR3') as catalog_setup:
    if os.environ.get('FERMI', None) is None:
        os.environ['FERMI'] ='.'
        print(f'Setting env var FERMI to ".". Expect to find folders catatag and skymodels')
    cat4 = Fermi4FGL()
    uwcat = UWcat(_uwname)

_sk = uwcat.skycoord.galactic
uwcat.loc[:,'glat'] = _sk.b.deg
uwcat.loc[:,'glon'] = _sk.l.deg
uwcat.loc[:, 'r95'] = np.sqrt( (uwcat.a * 2.64)**2 + (0.00791)**2) # 

src_finder = SourceFinder()

defaults = dict(neighbor=None, interval=30, nyquist=24, max_sep=0.5, tsmin=16, info_name='Other info',
            fft_query='p1>25 & 10>f>0.01', UTC=False)


# %% ../nbs/91_interface.ipynb 5
class Summarize():
    """ base class for call-back class"""

    def add(self, sa): 
        pass

class SourceAnalyzer():

    def __init__(self, name, 
            config=None,
            neighbor=None,  
            make_figs=True,
            summarizer=Summarize(),
            **kwargs):
        """
        """
        self.config = config if config is not None else Config()
        self.log=''
        self.wtl=None
        self.max_sep = kwargs.pop('max_sep', defaults['max_sep'])
        self.porb = float(kwargs.pop('porb', np.nan))
        self.fluxlim = kwargs.pop('fluxlim', None)
        interval = kwargs.pop('interval', defaults['interval'])
        nyquist  = kwargs.pop('nyquist', defaults['nyquist'])
        tsmin    = kwargs.pop('tsmin', defaults['tsmin'])
                              
        self.kwargs = kwargs 
        self.fig = None
        assert isinstance(summarizer, Summarize), f'Expect {summarizer} to be subclass of Summarize'
        
         
        SourceFinder.max_sep = self.max_sep #klugy
        if not self.setup(name, neighbor, interval, nyquist, tsmin): 
            return
        if make_figs:
            self.make_plots()
        try:
            summarizer.add(self)
        except Exception as msg:
            print(f"""SourceAnalyzer caught {summarizer.__class__.__name__}
            failure with source {self.name}: {msg} """, file=sys.stderr) 

    def plot_lc(self, ax1=None,ax2=None):
        if self.wtl is  None: return

        if ax1 is None:
            # OK: make figure with 2 subplots
            fig = plt.figure(figsize=(12,2.5))
            gs = plt.GridSpec(1, 2,  width_ratios=[4,1], wspace=0.2,hspace=0.5, top=0.95)
            (ax1,ax2) =  [fig.add_subplot(g) for g in gs]
        else: fig=None
        self.plot_bb(ax1)        
        self.plot_seds(ax2, self.max_sep)
        return fig

    def make_plots(self):

        if self.wtl is not None:
            self.fig=plt.figure(figsize=(12,5))
            gs = plt.GridSpec(2, 2,  width_ratios=[4,1], wspace=0.2,hspace=0.5, top=0.95)
            (ax1,ax2,ax3,ax4) =  [self.fig.add_subplot(g) for g in gs]
            self.plot_lc(ax1=ax1, ax2=ax2)
            self.plot_FFT(ax1=ax3, ax2=ax4)
        else:
            # no analysis since low TS. Only make the sed plots
            self.fig, ax = plt.subplots(figsize=(3,3))
            self.plot_seds(ax, self.max_sep)
        
    def __repr__(self):
        return f'{self.__class__.__name__}("{self.name}")'

    def setup(self, name, neighbor, interval, nyquist, tsmin, **kwargs):
        tstart = self.kwargs.get('tstart',0)
        tstop = self.kwargs.get('tstop',0)
        pt_name = self.pt_name =  src_finder.find(name, tol=self.max_sep)
        if pt_name is None: 
            self.log =  f'<font color="red">Failed lookup:</font> <br>{monospace(src_finder.log)} ' 
            #raise Exception(self.log)
            self.printout=''
            self.name='(not found)'
            self.skycoord = None
            return False
        self.skycoord = src_finder.skycoord
        if self.skycoord is None:
            raise Exception('Failed to locate source')
        self.name = name

        with capture_hide(f'Analysis printout for {name} ({pt_name})') as self.printout:
            print(src_finder.log)
            if pt_name not in uwcat.index:
                print(f'UW source {pt_name} not found in {uwname}??')
                self.wtl = self.bb=self.px = None
                self.pointlike_info = FileNotFoundError
                return False
            self.pointlike_info = pi = uwcat.loc[pt_name]
            ts = pi.ts
            if ts < tsmin:
                self.log = f'UW source has TS={ts:.1f} < {tsmin}'
                self.wtl = None
                return True

            self.wtl = WtLike(PointSource(pt_name, config=self.config), time_bins=(tstart, tstop, interval),) 

            if type(neighbor)==str and str(neighbor).strip():
                print(f'Reweighting with {neighbor}')
                self.neighbor_bb = WtLike(neighbor, config=self.config).bb_view()
                self.wtl = self.wtl.reweighted_view(self.neighbor_bb)
            self.bb = self.wtl.view(interval).bb_view()
            
            self.px = self.wtl.periodogram( 1/(4*nyquist), tstart=tstart, tstop=tstop)

            if not(np.isnan(self.porb)) and self.porb is not None:
                # generate orbial phase view to plot later (no periastron yet)
                bins = self.kwargs.get('phase_bins', 8)
                # create a phase view
                self.orbital_phase = self.wtl.phase_view(period=self.porb, nbins=bins)
                phase_df= self.orbital_phase.fits.query('t<1')
                self.orbital_ts = -2 * np.sum([f(1) for f in phase_df.fit.values])
        return True

    def get_catalog_info(self, cat, select=None, show=False):
        """
        select - list of column names
        """
        cone_size = SourceFinder.max_sep
        info = cat.catalog_entry( self.skycoord, cone_size=cone_size)
        if info is None:
            return f'No {cat.name} source within {cone_size} deg'
        name = info.name
        if select is not None:
            info = info[select]

        sep, r95 = info.sep, info.r95
        color = 'red' if sep>r95 else ''
        with capture(
                f'{cat.name} ({name}) <font color={color}>(sep, r95 ={sep*60:.1f}, {r95*60:.1f} arcmin )</font>',
                show=show) as out:
            print(info)
        return out

    def display_4fgl_info(self):
        try:
            return self.get_catalog_info(cat4,
            'fk5 galactic specfunc pivot eflux significance flags variability assoc_prob class1 assoc1_name  r95 sep'.split() )
        except:
            return 'No 4FGL-DR4 info'

    def display_pointlike_info(self):
        if self.pt_name is None:
            return ''
        if self.pt_name not in uwcat.index:
            return f'<font color="red"> Source {self.pt_name} not in {uwname}</font>'
        
        return self.get_catalog_info(uwcat,
            select='jname fk5 galactic specfunc e0 eflux100 ts fitqual locqual aprob r95  sep'.split(),
            )

    def _repr_html_(self):
        if self.fig is None:
            self.fig = make_plots()
            return FigureWrapper(self.fig)._repr_html_()
        return self.printout

    def plot_seds(self, ax=None, cone_size=0.5, **kwargs):
        _, ax = plt.subplots(figsize=(4,4)) if ax is None else (ax.figure, ax)
        c4entry = cat4.catalog_entry(self.skycoord, cone_size=cone_size)
        uwentry = uwcat.catalog_entry(self.skycoord, cone_size=cone_size)

        if uwentry is not None:
            uwentry.specfunc.sed_plot(ax=ax, e0=uwentry.e0, label=uwcat.name)
        if c4entry is not None:
            c4entry.specfunc.sed_plot(ax=ax, e0=c4entry.pivot, label=cat4.name)
            
        kw = dict( yticks=[0.1,1,10], xticks=[0.1,1,10],
            yticklabels='0.1 1 10'.split(), xticklabels='0.1 1 10'.split(),
                ylabel=r'$E^2\ dN/dE\ \mathrm{(eV\ cm^{-2}\ s^{-1})}$')
        kw.update(kwargs)
        ax.set(**kw)

    def get_harmonics(self, f0=None, p='p1', num=10):
        """ Return a DF with the num harmoics of f0
        """
        def htest(p):
            h = np.cumsum(p)-4*np.arange(len(p))
            if len(h)==0:
                return h, -1
            return h, np.argmax(h)
        
        f0 = 1/self.porb if f0 is None else f0
        pdf = self.px.power_df
        f= pdf.f.values
        fmax = f[-1]; imax= len(f)
        deltaf = fmax/imax
        if0 = int(f0/deltaf);  #index of fundamental
        nmax = imax//if0 # harmonic limit 
        # ret = np.full(num, np.nan)
        harmonics =  pdf.iloc[ if0 * np.arange(1, min(num+1, nmax))][p].values 
        h, imax = htest(harmonics)
        self.htest=h[imax] if imax>=0 else np.nan
        return harmonics, h, imax
    
    def get_bb_info(self):
        """ Return BB interval likelihood fits with 2-d (beta-free) fit also """
        
        def g2fit(cell):
            from wtlike.poisson import Poisson
            from wtlike.loglike import LogLike, Gaussian2dRep

            ts = Poisson.from_function(LogLike(cell)).ts
            r = dict(t=cell.t, tw=cell.tw, ts=round(ts,1))            
            if ts<4:
                r.update(flux=0, counts=cell.n, beta=np.nan, sig_beta=np.nan)
            else:
                r.update(Gaussian2dRep(LogLike(cell)).fit)
            return r
        try:
            df_bb = self.bb.fluxes['t tw ts flux errors'.split()]
            df_beta = pd.DataFrame.from_dict( 
                dict((i, g2fit(cell)) for i,cell in self.bb.cells.iterrows()) ,orient='index')\
                    ['flux beta sig_beta'.split()]
            return pd.concat([df_bb, df_beta], axis=1)

        except Exception as e:
            print(f'SourceAnalyzer: bb 2-d fit failed: {e}', file=sys.stderr)
            return None
        
    def plot_bb(self, ax, UTC=False):
        if self.fluxlim is not None:
            ax.set(ylim=self.fluxlim) ###
        self.bb.plot(ax=ax, source_name=self.wtl.source.name, UTC=UTC)
        self.bb_table = self.get_bb_info()

    def display_bb_table(self):
        """ return (hidden) markdown string"""
        bt = self.bb_table
        if bt is None or len(bt)==0 or not 'beta' in bt : return ''
        test = np.max(bt.beta/bt.sig_beta)
        
        warning_text = f'<font color=red>Check beta:  max beta/sig_beta is {test:.1f}</font>' if test>2 else ''
        # with capture_hide(f'BB fits {warning_text}') as cp:
        #     print(bt)
        # return cp
        dw = DataFrameWrapper(bt, summary=f'{len(bt)} BB fits {warning_text}', index=False)
        return dw._repr_html_()

    def plot_periodogram(self, ax3, ax4):
        
        def hist_peak_power(px, pname='p1', ax=None, xlim=(0,40), thresh=25,  **kwargs):
            """
            """
            p1 = px.find_peaks(pname).query('f>0.1')[pname].clip(*xlim)
            fig, ax = plt.subplots(figsize=(3,2)) if ax is None else (ax.figure, ax)
            kw = dict(xlabel=f'Power {pname}', xlim=xlim, title='' )
            kw.update(kwargs); ax.set(**kw)
            ax.hist(p1, np.linspace(*xlim,), log=True, histtype='step', lw=2);
            ax.hist(p1[p1>thresh], np.linspace(*xlim,), log=True, histtype='stepfilled', color='red');
            ax.grid(alpha=0.5)
            return fig
        # self.px = self.wtl.periodogram( 1/(4*nyquist) )
        self.px.power_plot(ax=ax3, pmax=50)
        hist_peak_power(self.px, ax=ax4, title='Peak distribution', xlabel='')

    def get_fft_peaks(self):
        if not hasattr(self, 'px'):
            return pd.DataFrame()
        query = defaults['fft_query']
        df = self.px.find_peaks('p1').query(query).copy()
        if len(df)>0: 
            df['period'] = 1/df.f
        return df
    
    def display_fft_peaks(self, show=False):
        try:
            df = self.get_fft_peaks()
        except:
            return monospace('FFt failed')
        query = defaults['fft_query']
        if len(df)==0:
            return monospace(f'No FFT peaks satisfying {query}')
   
        with pd.option_context('display.precision', 6, 'display.float_format',None):
            dw = DataFrameWrapper(df,
                    summary=f'{len(df)} FFT peaks satisfying {query}: max(p1)={max(df.p1):.1f}',
                    index=False, show=show)
            return dw._repr_html_()

    def get_nearby(self, radius_cut=3, var_cut=30):
        near_df = cat4.select_cone(self.skycoord, cone_size=5, query=f'variability>{var_cut}& sep<{radius_cut}')
        near_df['name'] = near_df.index
        return near_df['sep glon glat significance variability'.split()].sort_values('sep')

    def display_nearby(self, show=False, radius_cut=3, **kwargs):
        near_df = self.get_nearby(radius_cut=radius_cut, **kwargs)
        if len(near_df)==0: return f' No nearby (within {radius_cut} deg) variable sources.'
        dw = DataFrameWrapper(near_df,
            summary=f'{len(near_df) } nearby variable sources'
            )
        return dw._repr_html_()
        # with capture(f'{len(near_df) } nearby variable sources', show=show) as out:
        #     print(near_df)
        # return out

    def plot_FFT(self, ax1=None, ax2=None, peak_query='f>0.01'):
        """ Generate a Kerr Periodogram and a histogram of the peak values
        
        px -- TimeSeries object
        
        Return the figure with two Axes plots if had to create
        """
        def hist_peak_power(px, pname='p1', ax=None, xlim=(0,40), thresh=25,  **kwargs):
            """
            """
            p1 = px.find_peaks(pname).query(peak_query)[pname].clip(*xlim)
            _, ax = plt.subplots(figsize=(3,2)) if ax is None else (ax.figure, ax)
            kw = dict(xlabel=f'Power {pname}', xlim=xlim, title='' )
            kw.update(kwargs); ax.set(**kw)
            ax.hist(p1, np.linspace(*xlim,), log=True, histtype='step', lw=2);
            ax.hist(p1[p1>thresh], np.linspace(*xlim,), log=True, histtype='stepfilled', color='red');
            ax.grid(alpha=0.5)
        
        if ax1 is None:
            fig = plt.figure(figsize=(12,2.5))
            gs = plt.GridSpec(1, 2,  width_ratios=[4,1], wspace=0.2,hspace=0.5, top=0.95)
            (ax1,ax2) =  [fig.add_subplot(g) for g in gs]
        else: fig=None

        px = self.px
        px.power_plot(ax=ax1, pmax=50)
        hist_peak_power(px, ax=ax2,  xlabel='')
        ax2.set( xlabel=r'$P_1$ peak values')
        return fig

    def plot_orbital_phase(self, ax=None, **kwargs):
        if self.porb is None:
            print('No orbital period available', file=sys.stderr)
            return
       
        fig,ax = plt.subplots(figsize=(4,2)) if ax is None else (ax.figure, ax)
        kw = dict(xlim=(0,1), xticks=np.arange(0,1.01,0.25)); 
        kw.update(kwargs)
        self.orbital_phase.plot(ax=ax, **kw)
        ax.axvline( 1.0, ls='--', color='lightgrey')
        return fig
        
    def plot_phase(self, ax=None, **kwargs):

        period = kwargs.pop('period', self.porb)
        nbins  = kwargs.pop('nbins', 10)
        ref    = kwargs.pop('ref', '2008')        

        fig,ax = plt.subplots(figsize=(6,3)) if ax is None else (ax.figure, ax)
        kw = dict(xlim=(0,1),ylim=(0,None),); kw.update(kwargs)
        
        self.wtl.phase_view(period=period, nbins=nbins, reference=ref).plot(ax=ax, **kw)
        
        ax.axvline( 1.0, ls='--', color='lightgrey')
        return fig
    
    def plot_harmonics(self, ax=None, num=10, p='p1', **kwargs):

        # def htest(p):
        #     h = np.cumsum(p)-4*np.arange(len(p))
        #     return h, np.argmax(h)

        fig, ax = plt.subplots(figsize=(4,2)) if ax is None else (ax.figure, ax)
        harmonics, h, imax = self.get_harmonics(num=10)
        x = np.arange(1,len(harmonics)+1)
        ax.plot(x, harmonics, 'o--')
        ax.plot(x,  h, 'D--', color='orange')
        if imax>=0:
            ax.scatter(x[imax], h[imax], marker='*', c='red',s=200)
        t = ax.get_ylim()
        kw = dict(xlabel='Harmonic', ylabel='FFT power', ylim=(0, max(14.9,t[1])),
              xticks=range(1,11,2), xlim=(0.5, 10.5),)
        kw.update(kwargs)
        ax.set(**kw)
        ax.grid(alpha=0.5)
        return fig


# %% ../nbs/91_interface.ipynb 6
class SummaryInfo:  
    """Batch process a list of sources,
        returning a dict of summary info for each one
    """  
    
    def __call__(self, info, config=None):
        if type(info)==str:  name, info = info, {}
        else: name = info.name

        # run the analyzer without making figures
        try:
            sa = SourceAnalyzer(name, make_figs=False, tsmin=20, config=config)
        except Exception as e:
            return name, dict(error=str(e))
        if sa.name=='(not found)':
            return name, dict(error='source name not found')
        if sa is None:
            return name, dict(error='SourceAnalyzer failure')
        
        ret =  dict(
            light_curve = sa.get_bb_info(),
            fft_peaks = sa.get_fft_peaks(),
            nearby = sa.get_nearby(), 
            porb = sa.porb,)
        if not pd.isna(sa.porb):
            # items computed only porb is specified.
            ret.update(
                phase_lc = sa.orbital_phase.fits,
                harmonics = sa.get_harmonics(num=10) 
            )
        return name, ret
    
    def run(self, df):
        if isinstance(df, pd.DataFrame):
            # a data frame: column info passed to the analyzer
            return dict(df.apply(self, axis=1))
        elif hasattr(df, '__len__'):
            # a list of names
            return dict( map( lambda n: (n, self(n)),   df))
        elif type(df)==str:
            # perhas a 
            return self(df)
        else:
            raise Exception(f'Type {type(df)} not recognized')
        
def phase_plots(ldf, ncols=4):
    from wtlike.lightcurve import flux_plot
    nrows=(len(ldf)+ncols-1)//ncols
    fig, axx = plt.subplots(ncols=ncols, nrows=nrows, 
                            figsize=(12,nrows*2.+1), sharex=True)
    plt.subplots_adjust(wspace=0.2, hspace=0.05, top=0.935);
    axf = axx.flatten()
    for (name, info), ax in zip(ldf.items(), axf):
        pip = info['phase_lc']
        flux_plot(pip.iloc[:len(pip)//2], ax=ax, xlim=(0,1), legend_loc='none', 
                 step=True, colors=('cornflowerblue','cornflowerblue','blue'))
        ax.text(0.05,0.95, f'{name}\n{info["porb"]}', 
                transform=ax.transAxes, va='top', fontsize=10)
        t = ax.get_ylim()
        ax.set(ylabel='', xlabel='', ylim=(t[0], t[1]*1.1))
        ax.axhline(1, ls='--', color='grey')
    for ax in axf[len(ldf): len(axf)]: ax.set_visible(False)
    fig.suptitle('HMXB orbital phase plots', fontsize=18)
    return fig   

def harmonic_plots(ldf, ncols=4):
    nrows=(len(ldf)+ncols-1)//ncols
    fig, axx = plt.subplots(ncols=ncols, nrows=nrows, 
                            figsize=(12,nrows*2.+1), sharex=True)
    plt.subplots_adjust(wspace=0.2, hspace=0.1, top=0.935);
    axf = axx.flatten()
    for (name, info), ax in zip(ldf.items(), axf):
        pip, h,imax = info['harmonics']
        x = np.arange(1,len(pip)+1)
        ax.plot(x, pip, 'o--')

        ax.plot(x,  h, 'D--', color='orange')
        if imax>=0:
            ax.scatter(x[imax], h[imax], marker='*', c='red',s=200)
        ax.text(0.04,0.95, f'{name}', transform=ax.transAxes, va='top', fontsize=10)
        ax.text(0.97,0.95, f'{h[imax]:.0f}', 
                transform=ax.transAxes, va='top', ha='right', fontsize=10)
        t = ax.get_ylim()
        ax.set(ylabel='', xlabel='', ylim=(0, max(14.9,t[1]*1.3)),
              xticks=range(1,11,2), xlim=(0.5, 10.5))
        ax.grid(alpha=0.5)

    for ax in axf[len(ldf): len(axf)]: ax.set_visible(False)
    fig.suptitle('HMXB H-test', fontsize=18)
    return fig

# %% ../nbs/91_interface.ipynb 7
proc = None
def get_SA():
    """ access to the most recent SourceAnalyzer object"""
    return proc

@ipynb_doc
def examine_source(name, info=None, text='',  **kwargs): 

    """### {name}
    {other_info}
    {text}
    {log}
    {printout}
    {neighbor_plot}
    {pinfo}
    {ginfo}
    {fig}
    {nearby}
    {beta}
    {fft_peaks}
    {periodic_plots}
    
    """
    global proc
    ginfo=beta=fig=fft_peaks=nearby=neighbor_plot=pinfo=other_info=phase_plot=''
    kw = defaults.copy() 
    kw.update(kwargs)
    if info is not None: kw.update(info)
    max_sep = kw.pop('max_sep', None)
    info_name = kw.pop('info_name', 'Source info')
    if max_sep is not None: SourceFinder.max_sep=max_sep

    # the additional info, a dict-like object, can update default analysis parameters
    if info is not None:
        for k,v in info.items():
            if k in defaults and str(v)!='nan':
                # print(f'setting {k} from {kw[k]} to {v}')
                kw[k]=v

        with capture_hide(f'{info_name}') as other_info:
            print(info)
    else: other_info=''


    neighbor = kw.get('neighbor',None)
    if not str(neighbor).strip(): neighbor=None
    neighbor_info = '' if type(neighbor)!=str  else f'Reweighted with {neighbor}'
    
    try:
        proc = self = SourceAnalyzer(name, **kw)
    except Exception as ex:
        printout = f'<font color="red">Failed: {ex}</font>\n'
        raise
        return locals()
    
    log = '' if not self.log else f'<font color="red">{self.log}</font>'
    # proc = self = SourceAnalyzer(name, **kw)
    printout=self.printout
    if self.skycoord is None:
        return locals()
    pinfo = self.display_pointlike_info()
    ginfo = self.display_4fgl_info()
    fig = self.fig
    
    nearby = self.display_nearby()

    if self.wtl is not None:
        beta = self.display_bb_table()
        fft_peaks = self.display_fft_peaks()

    if type(neighbor) != str :
        neighbor_plot = ''
    elif self.wtl is not None: 
        neighbor_plot, ax = plt.subplots(figsize=(10,3))
        self.neighbor_bb.plot(ax=ax)
        neighbor_plot.summary= f'Reweighted with {neighbor} (click for its light curve)'

    if getattr(self, 'orbital_phase', None) is not None:
        periodic_plots, (ax1,ax2) = plt.subplots(ncols=2, figsize=(10,2.5))

        self.plot_orbital_phase(ax=ax1)
        self.plot_harmonics(ax=ax2)
        periodic_plots.summary =f"""Period {self.porb:.6f} d: TS {self.orbital_ts:.0f} 
        H-test {self.htest:.1f}"""
    else: periodic_plots=''
        

    return locals()

# %% ../nbs/91_interface.ipynb 8
class PhaseInfo(Summarize):
    
    def __init__(self, df):
        self.df = df        

    def add(self, sa): 
        """Called back from SourceAnalyzer
            Extract porb, the FFT power and phase view fits dataframes 
            into a dict
        """
        self.phasedict = dict(
                    powerdf = sa.px.power_df,
                    phasedf = sa.orbital_phase.fits,
                    porb = sa.porb,
                    )
        
    def __call__(self, idx):
        """ called to process self.df.iloc[idx]
        return (name, dict) tuple
        """
        info = self.df.iloc[idx]
        SourceAnalyzer(info.name, porb=info.porb, 
                    summarizer=self, make_figs=False, tsmin=20)  
        return  (info.name, self.phasedict)
    
    @classmethod
    def multirun(cls, df, poolsize=4, cache_file=None, update=False):
        """Run a set of sources in parallel, using SourceAnalyzer
        df has name and porb 
        """
        from wtlike.config import Timer
        from multiprocessing import Pool
        import pickle
        
        cf = Path(cache_file)
        if cf.is_file() and not update:
            with open(cf, 'rb') as inp:
                ldf =pickle.load(inp)
                show(f'Read cache file {cf} with {len(ldf)} sources')
                return ldf
                     
        N = len(df)
        show(f'### Analyzing {N} sources with {poolsize} processes...')
        t=Timer()
        if poolsize==1:
            ldf = list(map(cls(df), range(N)))
        else:
            with Pool(processes=poolsize) as pool:
                ldf = pool.map(cls(df), range(N)) 
        show(t)
        if cf is not None:
            with open(cf, 'wb') as out:
                pickle.dump(ldf, out)
            show(f'Cache file saved to {cf}')
        return ldf


# %% ../nbs/91_interface.ipynb 9
def process_df(df, **kwargs): #, max_sep=None, tsmin=50):

    defaults.update(kwargs)
    with capture('Default parameter values') as parout:
        print(pd.Series(defaults))

    show(f"""\
        ## Processing {len(df)} sources, start {str(datetime.datetime.now())[:16]}
        {parout}
        
        ---
        """)
    for name, info in df.iterrows():
        if type(name) !=str:
            print(f'Bad entry: {name}', file=sys.stderr)
            continue
        examine_source(name, info=info, **kwargs ) 

    show(f"""\
    
        \n---
        \n## Finish at {str(datetime.datetime.now())[:16]}""")

class WTSkyCoord(SkyCoord):
    def __repr__(self):
        ra,dec = self.fk5.ra.deg, self.fk5.dec.deg
        return f'({ra:7.3f},{dec:+7.3f})'


def load_source_spreadsheet(
            filename='AMXP scorecard (2).xlsx',
            source_name='AMXP name',
            ra_name=None, dec_name=None,
            flag_row=True):
    """Return a dataframe derived from the table in a spreadsheet
    row 1 is column names
    if flag_row is True row 2 has non-blank to include the column in the datafrane
    if ra_name (and dec_name) are not None, rename columns to "ra" and "dec"
    The index is set to the contents of source_nane
    """
    spreadsheet = Path(filename) #'AMXP scorecard (2).xlsx')
    assert spreadsheet.is_file(), f'File "{filename}" not found'
    df = pd.read_excel(spreadsheet)
    assert source_name in df.columns, f'did not find column "{source_name}" to use as index'
    # take care of trailing non-break space
    df.index = list(map(lambda s:str(s).strip().replace(u"\u2014","-"),df.loc[:,source_name]))
    if flag_row:
        to_drop = pd.isna(df.iloc[0])
        to_drop[source_name]=True
        df = df.drop(columns=df.columns[to_drop])[1:]
    else:
        df = df.drop(columns=[source_name])
    if ra_name is not None:
        df.rename(columns={ra_name: 'ra', dec_name: 'dec'}, inplace=True)
    return df

def get_fermi_info(source_names, max_sep=0.5):
    """
    Return a dataframe with information on UW and 4FGL sources near the locations of a list of source names
    """
    
    src_finder = SourceFinder()
    SourceFinder.max_sep = max_sep
    
    class SourceInfo(dict):

        # class SkyCoord(SkyCoord):
        #     # subclass that displays (ra,dec)
        #     def __repr__(self):
        #         ra,dec = self.fk5.ra.deg, self.fk5.dec.deg
        #         return f'({ra:6.3f},{dec:+6.3f})'

        def __init__(self, name):
            #self.name=name.replace('_',' ')
            src_finder.log=''

            # get the nearest pointlike name from the weight tables
            self.pt_name = pt_name = src_finder(name) 

            try:
                self['skycoord'] = WTSkyCoord(src_finder.skycoord)
            except ValueError as err:
                print(f'SourceInfo: Fail to recognize name "{name}"', file=sys.stderr)
                return

            if pt_name is None: return

            # grab pointlike info
            uw = self.get_uwinfo(pt_name)
            if len(uw)==0: 
                self.update(uw_name = pt_name) 
                return
            self.update(uw_name=pt_name,
                        uw_pos=WTSkyCoord(uw.ra,uw.dec,unit='deg',frame='fk5'), 
                        uw_sep = uw.sep,
                        uw_r95 = uw.r95,
                        uw_ts = uw.ts,
                        uw_sed= uw.specfunc,
                        uw_jname = uw.jname,
                       )
            dr = self.get_4fgl()
            if len(dr)==0: return
            self.update(dr_name=dr.name,
                        dr_pos=WTSkyCoord(dr.ra,dr.dec,unit='deg', frame='fk5'),
                        dr_sep = dr.sep,
                        dr_r95 = dr.r95,
                        dr_class1=dr.class1)
        def get_uwinfo(self, pt_name):
            #return a dict 
            if pt_name not in uwcat.index:
                # print(f'{self.name}: UW source {pt_name} not found in {uwname}??', file=sys.stderr)
                return pd.Series(name=pt_name, dtype=object)
            info = uwcat.catalog_entry( self['skycoord'], cone_size=SourceFinder.max_sep)
            return pd.Series(info)

        def get_4fgl(self):
            info = cat4.catalog_entry( self['skycoord'], cone_size=SourceFinder.max_sep)
            if info is None: return pd.Series(dtype=object)
            return pd.Series(info)

        def __repr__(self):
            return f'{self.__class__.__name__}({self.name}):\n{pd.Series(self, dtype=object)}'
    
    source_names = np.atleast_1d(source_names)
    return pd.DataFrame([SourceInfo(name) for name in source_names], index=source_names )

# %% ../nbs/91_interface.ipynb 10
def setup_excel(filename, 
        source_name_column,
        ra_name=None, dec_name=None,
        title='Excel analysis',
        query='',
        select = (None,),
        **kwargs):

    defaults.update(kwargs)

    spreadsheet = Path(filename) #'AMXP scorecard (2).xlsx')
    assert spreadsheet.is_file(), f'File "{filename}" not found'
    if len(title)>1 and title[0] != '#' : title = '# '+title
    show(f"""\
        {title}

        {catalog_setup}
        Read spreadsheet "{spreadsheet.absolute()}", 
        dated {str(datetime.datetime.fromtimestamp(spreadsheet.stat().st_mtime))[:16]}
        """)
    # make the index the source name, remove columns without a "1" (or anything?) in the second row
    df = pd.read_excel(spreadsheet)
    assert source_name_column in df.columns, f'did not find column "{source_name_column}" to use as index'
    df.index = df.loc[:,source_name_column]
    to_drop = pd.isna(df.iloc[0])
    if type(select) != tuple: 
        # assume selecting a single row
        select = (select,select+1)
    df = df.drop(columns=df.columns[to_drop])[1:][slice(*select)]
    if ra_name is not None and dec_name is not None:
        sc = SkyCoord(df[ra_name], df[dec_name], unit='deg', frame='fk5')
        df.loc[:,'skycoord'] = CatDF.SkyCoord(sc)
    if query:        
        df = df.query(query)
    return df

def process_excel(filename, source_name_column,  **kwargs):
    df = setup_excel(filename, source_name_column,  **kwargs)
    return process_df(df, **kwargs)
